package net.sourceforge.floatapi;

import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Window;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.dnd.DragSource;
import java.awt.dnd.DragSourceAdapter;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DropTargetListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.JComponent;



/*
 * Version 1.0
 *
 * This Class is Open Source Code under LGPL License.
 *
 *  copyright by floatAPI project group
 * contact us over http://floatAPI.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Date: 12.10.2003  Time: 14:06:40
 *
 */

/**
 * <code>FloatingHandler</code> makes floating of added <code>Floating</code>s and <code>Dockable</code>s possible.
 *
 * A <code>FloatingHandler</code> may be constructed for a special domain in a user interface or a
 * default static <code>FloatingHandler</code> may be obtained through {@link #getDefaultHandler()}.
 * During floating events a defined gestureCursor is used as Cursor (default is Cursor.DEFAULT_CURSOR).
 * <p> To enable floating of <code>Floating</code>s over <code>Dockable</code>s it is vital, that all involved
 * <code>Floating</code>s and <code>Dockable</code>s are registered. It is good habit in terms of garbage collection, but not strictly necessary
 * to unregister these after usage.
 *
 * <PRE>
 * //example implementation for floating use:
 *
 * //get the default FloatingHandler
 * FloatingHandler floatingHanlder=FloatingHandler.getDefaultHandler();
 *
 * //setting a special Cursor for floating gestures
 * floatingHandler.setGestureCursor(new Cursor(Cursor.HAND_CURSOR);
 *
 * //register the participating objects
 * floatingHandler.add(dockable);
 * floatingHandler.add(floating);
 *
 * //programatically let floating be docked in dockable
 * floatingHandler.dock(dockable,dockable);
 *
 * // ... after some special event on a button for example ...
 * //programatically make floating float (out of the current dockable) to a location on the screen
 * floatingHandler.floatTo(floating,new Point(10,10));
 *
 * //unregister the participating objects to release the reference
 * floatingHandler.remove(dockable);
 * floatingHandler.remove(floating);
 * </PRE>
 *
 * @see Floating
 * @see Dockable
 * @see JFloater
 * @see JDocker
 */
public class FloatingHandler {
    private static final int acceptDNDs = DnDConstants.ACTION_REFERENCE;

	private static volatile FloatingHandler myself;
	private static int idCount = 0;

    private DataFlavor FLOATING_FLAVOR = FloatingFlavor.createFlavor();

	private ArrayList floaters;
	private ArrayList dockers;
	private HashMap handles;
	private HashMap floatings;

	private Cursor gestureCursor=new Cursor(Cursor.DEFAULT_CURSOR);
	private String unique = "unique"+Math.random();

    /**
     * constructs a (user-interface domain specific) <code>FloatingHandler</code>
     */
    public FloatingHandler() {
        floaters = new ArrayList();
        dockers = new ArrayList();
        handles = new HashMap();
        floatings = new HashMap();
    }

    /**
     * @return the static reference to the default <code>FloatingHandler</code>
     */
    public static FloatingHandler getDefaultHandler() {
        if (myself == null){
        	synchronized(FloatingHandler.class){
        		if (myself == null){
					myself = new FloatingHandler();
        		}
        	}
        }            
        return myself;
    }

    /**
     * @return <code>Cursor</code> defined to use during floating gestures
     */
    public Cursor getGestureCursor() {
        return gestureCursor;
    }

    /**
     * define the <code>Cursor</code> to use during floating gestures
     *
     * @param gestureCursor to be used during floating gestures
     */
    public void setGestureCursor(final Cursor gestureCursor) {
        this.gestureCursor = gestureCursor;
    }

    /**
     * registers the floating object and provides floating functionality for it
     *
     * @param floating to be registered
     */
    public void add(final Floating floating) {
        String handle = floating.getFloatingType() + "_[" + (idCount++) + ']';
        FloaterHandle f = new FloaterHandle(handle, floating);
        floaters.add(f);
        handles.put(handle, f);
        floatings.put(floating, f);
    }

    /**
     * unregisters the floating object and releases internal references to it.
     *
     * @param floating to be unregistered and to lose floating functionality
     * @return true if successful, false if floating was not registered
     */
    public boolean remove(final Floating floating) {
        FloaterHandle f = (FloaterHandle) floatings.get(floating);
        if (f == null)
            return false;
        f.release();
        floaters.remove(f);
        handles.remove(f.handle);
        floatings.remove(f.floating);
        return true;
    }

    /**
     * registers the dockable object and provides dockable functionality on it
     *
     * @param dockable to be registered
     */
    public void add(final Dockable dockable) {
        dockers.add(new DockerHandle(dockable));
    }

    /**
     * unregisters the dockable object and releases internal references to it.
     *
     * @param dockable to be unregistered and to lose dockable functionality
     * @return true if successful, false if dockable was not registered
     */
    public boolean remove(final Dockable dockable) {
        Iterator it = dockers.iterator();
        while (it.hasNext()) {
            DockerHandle d = (DockerHandle) it.next();
            if (d.dockable == dockable) {
                d.release();
                it.remove();
                return true;
            }
        }
        return false;
    }

    /**
     * make floating float (out of the current dockable) to a location on the screen
     *
     * @param floating to make floating
     * @param p location on screen
     */
    public void floatTo(final Floating floating, final Point p) {
    	Dockable dockable = floating.getFloater();
        Window window = getTop(dockable.getDockTarget());
        Component component = floating.getFloatingComponent();
        Dimension d = component.getSize();                    
		component.setSize(d);
		if (component instanceof JComponent){
			((JComponent)component).setPreferredSize(d);
			((JComponent)component).setMinimumSize(d);
			((JComponent)component).setMaximumSize(d);			
		}
        window.setLocation(p);
        dock(floating, dockable);
		window.pack();
        window.setVisible(true);
    }
    
    /**
     * let floating be docked in destination
     *
     * @param floating to dock in destination
     * @param destination where floating is to be docked in
     * @return if successfully docked floating in destination
     */
    public boolean dock(final Floating floating, final Dockable destination) {
        if (floating.getCurrentDock() != destination) {
            if (destination.isDockable(floating)) {
                Container source = (floating.getCurrentDock() != null) ? floating.getCurrentDock().getDockTarget() : null;
                if (source != null) {
					floating.getCurrentDock().undock(floating);
                }

                destination.dock(floating);
                floating.setCurrentDock(destination);

                if (source != null) {
                    source.doLayout();
                    if (source instanceof JComponent) {
                        ((JComponent) source).revalidate();
                    }
                    source.repaint();
                }

                destination.getDockTarget().doLayout();
                if (destination instanceof JComponent) {
                    ((JComponent) destination.getDockTarget()).revalidate();
                }
                destination.getDockTarget().repaint();
                return true;
            }
        }
        return false;
    }
    
    /**
     * programmatically start dragging of a floating.
     *
     * This method can be used to trigger dragging gesture through a (not necessarily involved with 
     * this actual floating) <code>InputEvent</code>.   
     * The <code>InputEvent</code> should be such as of a pressed or dragged <code>MouseEvent</code> 
     * in order to allow a drop to occure with a following released <code>MouseEvent</code>.
     * @param floating to start the initial dragging gesture on .
     * @param inputEvent used to trigger dragging gesture on the floating.
     */
    public void startDrag(Floating floating, InputEvent inputEvent){
    	FloaterHandle handle = (FloaterHandle)floatings.get(floating);
    	if (handle != null) {
    		if (handle.dragSources.size() > 0){
				DragSource dragSource = (DragSource)handle.dragSources.get(0);
				Component componentTrigger = floating.getFloatingTriggers()[0];
				VirtualDragGestureRecognizer mdgr = 
					new VirtualDragGestureRecognizer(dragSource, componentTrigger, acceptDNDs, null);	
				mdgr.resetRecognizer();
				mdgr.appendEvent(inputEvent);

				ArrayList list = new ArrayList();
				list.add(inputEvent);
				Point p;
				if (inputEvent instanceof MouseEvent){
					p = ((MouseEvent)inputEvent).getPoint();
				} else{
					p = new Point(0,0);
				}
				DragGestureEvent dge = new DragGestureEvent(mdgr, acceptDNDs, p, list);
				handle.dragGestureRecognized(dge);    			    		
    		}
    	}
    }

    public static Window getTop(final Component c) {
        Component current = (c instanceof JComponent) ? ((JComponent) c).getRootPane() : c;
        while (current != null) {
            current = current.getParent();
            if (current instanceof Window)
                return (Window) current;
        }
        return null;
    }

    class FloaterHandle extends DragSourceAdapter implements DragGestureListener, Transferable {
        Floating floating;
        String handle;
        Point currentOrigin;
        ArrayList dragSources = new ArrayList();
        ArrayList gestureRecognizers = new ArrayList();

        public FloaterHandle(String handle, Floating floating) {
            this.handle = handle;
            this.floating = floating;
            //register floating
            Component[] cs = floating.getFloatingTriggers();
            for (int i = 0; i < cs.length; i++) {
                Component c = cs[i];
                DragSource dragSource = new DragSource();
                DragGestureRecognizer gestureRecognizer = dragSource.createDefaultDragGestureRecognizer(c, acceptDNDs, this);
                if (gestureRecognizer == null)
                    throw new IllegalStateException("ACTION_REFERENCE DragGestureRecognizer could not be created for " + c);
                dragSource.addDragSourceListener(this);
                dragSources.add(dragSource);
                gestureRecognizers.add(gestureRecognizer);
            }

        }

        public void release() {
            //unregister floating
            Iterator it = gestureRecognizers.iterator();
            while (it.hasNext()) {
                DragGestureRecognizer gestureRecognizer = (DragGestureRecognizer) it.next();
                gestureRecognizer.resetRecognizer();
            }
            it = dragSources.iterator();
            while (it.hasNext()) {
                DragSource dragSource = (DragSource) it.next();
                dragSource.removeDragSourceListener(this);
            }

            floating = null;
            handle = null;
        }

        public void dragDropEnd(DragSourceDropEvent dsde) {
            if (!dsde.getDropSuccess()) {
                Point p = dsde.getLocation();
                p.x -= currentOrigin.x;
                p.y -= currentOrigin.y;
                floatTo(floating, p);
            }
        }

        public void dragGestureRecognized(DragGestureEvent dge) {
            dge.startDrag(getGestureCursor(), this);
            currentOrigin = dge.getDragOrigin();
        }

        public DataFlavor[] getTransferDataFlavors() {
            DataFlavor[] df = new DataFlavor[1];
            df[0] = FLOATING_FLAVOR;
            return df;
        }

        public boolean isDataFlavorSupported(DataFlavor flavor) {
            if (flavor == FLOATING_FLAVOR) {
                return true;
            }
            return false;
        }

        public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException {
            if (flavor == FLOATING_FLAVOR) {
                return new FloatingFlavor(handle, unique);
            }
            throw new UnsupportedFlavorException(flavor);
        }
    }

    class DockerHandle implements DropTargetListener {
        Dockable dockable;
        DropTarget dt;

        public DockerHandle(Dockable dockable) {
            this.dockable = dockable;
            //register dockable
            dt = new DropTarget(dockable.getDockTarget(), this);
            dt.setActive(true);

        }

        public void release() {
            //unregister dockable
            dt.setActive(false);
            dt.setComponent(null);
            dt.removeDropTargetListener(this);
            dockable = null;
        }

        public void dragEnter(DropTargetDragEvent dtde) {
            DataFlavor[] dfs = dtde.getCurrentDataFlavors();
            for (int i = 0; i < dfs.length; i++) {
                DataFlavor df = dfs[i];
                if (df == FLOATING_FLAVOR) {
                    dtde.acceptDrag(acceptDNDs);

                    dockable.updateDockGesture(true);
                    return;
                }
            }
            dtde.rejectDrag();
            dockable.updateDockGesture(false);
        }

        public void dragExit(DropTargetEvent dte) {
            dockable.updateDockGesture(false);
        }

        public void dragOver(DropTargetDragEvent dtde) {
        }

        public void drop(DropTargetDropEvent dtde) {
            DataFlavor[] dfs = dtde.getCurrentDataFlavors();
            for (int i = 0; i < dfs.length; i++) {
                DataFlavor df = dfs[i];
                if (df == FLOATING_FLAVOR) {
                    Transferable transferable = dtde.getTransferable();
                    if (transferable.isDataFlavorSupported(FLOATING_FLAVOR)) {
                        dtde.acceptDrop(acceptDNDs);

                        Floating floating = null;

                        try {
                            FloatingFlavor floatingFlavor=(FloatingFlavor) transferable.getTransferData(FLOATING_FLAVOR);
                            String handle = floatingFlavor.getHandle();
                            String unique=floatingFlavor.getUnique();

                            if (!FloatingHandler.this.unique.equals(unique)){
                                dockable.updateDockGesture(false);
                                dtde.dropComplete(true);
                                return;
                            }

                            FloaterHandle floaterHandle = (FloaterHandle) handles.get(handle);
                            if (floaterHandle==null){
                                dockable.updateDockGesture(false);
                                dtde.dropComplete(true);
                                return;
                            }

                            floating = floaterHandle.floating;
                        } catch (UnsupportedFlavorException e) {
                            e.printStackTrace();
                        } catch (IOException e) {
                            e.printStackTrace();
                        }
                        dockable.updateDockGesture(false);

                        if (floating != null) {
                            dock(floating, dockable);
                        }

                        dtde.dropComplete(true);
                    }
                    return;
                }
            }
            dtde.rejectDrop();
            dockable.updateDockGesture(false);
        }

        public void dropActionChanged(DropTargetDragEvent dtde) {
            DataFlavor[] dfs = dtde.getCurrentDataFlavors();
            for (int i = 0; i < dfs.length; i++) {
                DataFlavor df = dfs[i];
                if (df == FLOATING_FLAVOR) {
                    dtde.acceptDrag(acceptDNDs);

                    dockable.updateDockGesture(true);

                    return;
                }
            }
            dtde.rejectDrag();
            dockable.updateDockGesture(false);
        }
    }

	class VirtualDragGestureRecognizer extends DragGestureRecognizer{
		public VirtualDragGestureRecognizer(DragSource ds, Component c, int sa, DragGestureListener dgl) {
			super(ds,c,sa,dgl);
		} 
		
		public void registerListeners(){ }
		
		public void unregisterListeners(){ }
					
		public synchronized void appendEvent(InputEvent awtie) {
			super.appendEvent(awtie);
		}								
};

}