package net.sourceforge.floatapi;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Window;

import javax.swing.JPanel;


/*
 * Version 1.0
 *
 * This Class is Open Source Code under LGPL License.
 *
 *  copyright by floatAPI project group
 * contact us over http://floatAPI.sourceforge.net
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Date: 12.10.2003  Time: 14:05:23
 *
 */

/**
 * <code>JDocker</code> is a convenience <code>JPanel</code> implementing <code>Dockable</code>.
 *
 * <code>JDocker</code> should not be used as <code>Container</code>s for other <code>Component</code> as its
 * children are manipulated during docking process. Furthermore a restrictToFloatingType can be set which will instruct
 * this <code>JDocker</code> to only accept docking of <code>Floatings</code> with the equivalent type.
 *
 */
public class JDocker extends JPanel implements Dockable {
    boolean arriving = false;
    boolean isFloater = false;
    Color highLightColor;
    String restrictToFloatingType;	

    /**
     * constructs a <code>JDocker</code>.
     *
     * a <code>BorderLayout</code> is set as layout manager.
     * @param isFloater <code>true</code>, if this <code>JDocker</code> 
     *        should be set invisible, if no <code>Floating</code> is docked.
     */
    public JDocker(boolean isFloater){
      setLayout(new BorderLayout());
      this.isFloater = isFloater;
    }

    /**
     * construct a <code>JDocker</code> which is not a Floater (stays visible even if no 
     * <code>Floating</code> is docked) and with the given restriction of type of 
     * <code>Floating</code> to provide docking for.
     *
     * a <code>BorderLayout</code> is set as layout manager.
     * @param restrictToFloaterType the type or <code>null</code>, if any
     */
    public JDocker(String restrictToFloaterType) {
        this(false);
        setRestrictToFloatingType(restrictToFloaterType);
    }

    /**
     * retrieve the currently type of <code>Floating</code> to provide docking for.
     * @return the type or <code>null</code>, if any
     */
    public String getRestrictToFloatingType() {
        return restrictToFloatingType;
    }

    /**
     * define which type of <code>Floating</code> to provide docking for.
     *
     * @param restrictToFloatingType the type or <code>null</code>, if any
     */
    public void setRestrictToFloatingType(String restrictToFloatingType) {
        this.restrictToFloatingType = restrictToFloatingType;
    }

    /**     
     * @see Dockable#getDockTarget()
     * @return <code>this</code>
     */
    public Container getDockTarget(){
      return this;
    }

    /**
     * @see Dockable#updateDockGesture(boolean)
     * @param arriving true if, false if not
     */
    public void updateDockGesture(boolean arriving){
        this.arriving=arriving;
        repaint();
    }

    /**
     * overrides by calling super, then possibly highLighting the component.
     *
     * A through {@link #updateDockGesture(boolean)} triggered event is graphically represented by filling the complete
     * bounds of this JDocker with highLightColor or if undefined, with half transparent foreground Color.
     *
     * @param g <code>Graphics</code> currently painting to
     */
    public void paintComponent(Graphics g) {
        super.paintComponent(g);

        if (arriving) {
            Color c = highLightColor;
            if (c == null){
                Color fg=getForeground();
                c = new Color(fg.getRed(),fg.getGreen(),fg.getBlue(),128);
            }
            g.setColor(c);
            ((Graphics2D) g).fill(getBounds());
        }
    }

    /**
     * This implementation checks:
     * <UL>
     * <LI>if restrictToFloatingType is defined, only a floating of the equivalent type may dock.
     * <LI>if there is already a <code>Floating</code> component docked.
     * </UL>
     *
     * @see Dockable#isDockable(Floating)
     * @param floating to dock
     * @return true if docking of floating is possible, false if not
     */
    public boolean isDockable(Floating floating) {
        if (restrictToFloatingType!=null){
            if (!restrictToFloatingType.equals(floating.getFloatingType())){
                return false;
            }
        }
        if (getComponentCount() == 0) return true;
        for (int i = 0; i < getComponentCount(); i++) {
            Component c = getComponent(i);
            if (c instanceof Floating) return false;
        }
        return true;
    }

    /**
     * @see Dockable#getDocked()
     * @return the currently docked <code>Floating</code> or null, if vacant.
     */
    public Floating getDocked(){
        if (getComponentCount() == 0) return null;
        for (int i = 0; i < getComponentCount(); i++) {
            Component c = getComponent(i);
            if (c instanceof Floating) return (Floating) c;
        }
        return null;

    }

    /**
     * removes all currently contained <code>Component</code>s and adds the floating at <code>BorderLayout.CENTER</code>.
     *
     * @see Dockable#dock(Floating)
     * @param floating to dock in this JDocker
     */
    public void dock(Floating floating){
        removeAll();
        add(floating.getFloatingComponent(), BorderLayout.CENTER);
    }
    
	/**
	 * fulfill the actual act of undocking floating from this <code>Dockable</code>
	 *
	 * @param floating to undock from this
	 */
	public void undock(Floating floating){
		if (isFloater){
			Window lastWindow = FloatingHandler.getTop(getDockTarget());
			lastWindow.setVisible(false);

			getDockTarget().remove(floating.getFloatingComponent());

		}
	}
}